/*
 *    Instance.java
 *    Copyright (C) 1999 Eibe Frank
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package weka.core;

import java.util.*;
import java.io.*;

/**
 * Class for handling an instance. All values (numeric, nominal, or
 * string) are internally stored as floating-point numbers. If an
 * attribute is nominal (or a string), the stored value is the index
 * of the corresponding nominal (or string) value in the attribute's
 * definition. We have chosen this approach in favor of a more elegant
 * object-oriented approach because it is much faster. <p>
 *
 * Typical usage (code from the main() method of this class): <p>
 *
 * <code>
 * ... <br>
 *      
 * // Create empty instance with three attribute values <br>
 * Instance inst = new Instance(3); <br><br>
 *     
 * // Set instance's values for the attributes "length", "weight", and "position"<br>
 * inst.setValue(length, 5.3); <br>
 * inst.setValue(weight, 300); <br>
 * inst.setValue(position, "first"); <br><br>
 *   
 * // Set instance's dataset to be the dataset "race" <br>
 * inst.setDataset(race); <br><br>
 *   
 * // Print the instance <br>
 * System.out.println("The instance: " + inst); <br>
 *
 * ... <br>
 * </code><p>
 *
 * All methods that change an instance are safe, ie. a change of an
 * instance does not affect any other instances. All methods that
 * change an instance's attribute values clone the attribute value
 * vector before it is changed.
 *
 * @author Eibe Frank (eibe@cs.waikato.ac.nz)
 * @version $Revision: 1.3 $ 
*/
public class Instance implements Copyable, Serializable, InstanceContainer {
  
  /** Constant representing a missing value. */
  private final static double MISSING_VALUE = Double.NaN;

  /** 
   * The dataset the instance has access to.  Null if the instance
   * doesn't have access to any dataset.  Only if an instance has
   * access to a dataset, it knows about the actual attribute types.  
   */
  protected Instances m_Dataset;

  /** The instance's attribute values. */
  protected double[] m_AttValues;

  /** The instance's weight. */
  protected double m_Weight;
    
  protected int m_index = -1; // in dataset if exists
  
  public Instance() {    
    m_AttValues = null;
    m_Weight = 1;
    m_Dataset = null;             		
  }
  /************************************************/
  
  /**
   * Constructor that copies the attribute values and the weight from
   * the given instance. Reference to the dataset is set to null.
   * (ie. the instance doesn't have access to information about the
   * attribute types)
   *
   * @param instance the instance from which the attribute
   * values and the weight are to be copied 
   */

  /**
   * Constructor that inititalizes instance variable with given
   * values. Reference to the dataset is set to null. (ie. the instance
   * doesn't have access to information about the attribute types)
   *
   * @param weight the instance's weight
   * @param attValues a vector of attribute values 
   */
  public Instance(double weight, double[] attValues){
    
    m_AttValues = attValues;
    m_Weight = weight;
    m_Dataset = null;    
  }

  /**
   * Constructor of an instance that sets weight to one, all values to
   * be missing, and the reference to the dataset to null. (ie. the instance
   * doesn't have access to information about the attribute types)
   *
   * @param numAttributes the size of the instance 
   */
  public Instance(int numAttributes) {
    
    m_AttValues = new double[numAttributes];
    for (int i = 0; i < m_AttValues.length; i++) {
      m_AttValues[i] = MISSING_VALUE;
    }
    m_Weight = 1;
    m_Dataset = null;
  }

  /**
   * Returns the attribute with the given index.
   *
   * @param index the attribute's index
   * @return the attribute at the given position
   * @exception Exception if instance doesn't have access to a
   * dataset
   */ 
  public Attribute attribute(int index) throws Exception {
   
    if (m_Dataset == null) {
      throw new Exception("Instance doesn't have access to a dataset!");
    }
    return m_Dataset.attribute(index);
  }

  /**
   * Returns class attribute.
   *
   * @return the class attribute
   * @exception Exception if the class is not set or the
   * instance doesn't have access to a dataset
   */
  public Attribute classAttribute() throws Exception {

    if (m_Dataset == null) {
      throw new Exception("Instance doesn't have access to a dataset!");
    }
    return m_Dataset.classAttribute();
  }

  /**
   * Returns the class attribute's index.
   *
   * @return the class index as an integer 
   * @exception Exception if instance doesn't have access to a dataset 
   */
  public int classIndex() throws Exception {
    
    if (m_Dataset == null) {
      throw new Exception("Instance doesn't have access to a dataset!");
    }
    return m_Dataset.classIndex();
  }

  /**
   * Tests if an instance's class is missing.
   *
   * @return true if the instance's class is missing
   * @exception Exception if the class is not set or the instance doesn't
   * have access to a dataset
   */
  public boolean classIsMissing() throws Exception {

    if (classIndex() < 0) {
      throw new Exception("Class is not set!");
    }
    return isMissing(classIndex());
  }

  /**
   * Returns an instance's class value in internal format. (ie. as a
   * floating-point number)
   *
   * @return the corresponding value as a double (If the 
   * corresponding attribute is nominal (or a string) then it returns the 
   * value's index as a double).
   * @exception Exception if the class is not set or the instance doesn't
   * have access to a dataset 
   */
  public double classValue() throws Exception {
    
    if (classIndex() < 0) {
      throw new Exception("Class is not set!");
    }
    return value(classIndex());
  }

  /**
   * Produces a shallow copy of this instance. The copy doesn't have
   * access to a dataset.
   *
   * @return the shallow copy
   */
  public Object copy() {
    return this.makeCopy();    
    //return new Instance(this);
  }

  public Instance makeCopy()
  {
    Instance ins = new Instance(this.m_Weight , this.m_AttValues);
    ins.setIndex(this.m_index);
    ins.setDataset(this.m_Dataset);
    return ins;    
  }
  /**
   * Returns the dataset this instance has access to. (ie. obtains
   * information about attribute types from) Null if the instance
   * doesn't have access to a dataset.
   *
   * @return the dataset the instance has accesss to
   */
  public Instances dataset() {

    return m_Dataset;
  }

  /**
   * Deletes an attribute at the given position (0 to 
   * numAttributes() - 1). Only succeeds if the instance does not
   * have access to any dataset because otherwise inconsistencies
   * could be introduced.
   *
   * @param pos the attribute's position
   * @exception Exception if the instance has access to a
   * dataset 
   */
  public void deleteAttributeAt(int position) 
       throws Exception {

    if (m_Dataset != null) {
      throw new Exception("Instance has access to a dataset!");
    }
    forceDeleteAttributeAt(position);
  }

  /**
   * Returns an enumeration of all the attributes.
   *
   * @return enumeration of all the attributes
   * @exception Exception if the instance doesn't
   * have access to a dataset 
   */
  public Enumeration enumerateAttributes() throws Exception {

    if (m_Dataset == null) {
      throw new Exception("Instance doesn't have access to a dataset!");
    }
    return m_Dataset.enumerateAttributes();
  }

  /**
   * Tests if the headers of two instances are equivalent.
   *
   * @param instance another instance
   * @return true if the header of the given instance is 
   * equivalent to this instance's header
   * @exception Exception if instance doesn't have access to any
   * dataset
   */
  public boolean equalHeaders(Instance inst) 
       throws Exception {

    if (m_Dataset == null) {
      throw new Exception("Instance doesn't have access to a dataset!");
    }
    return m_Dataset.equalHeaders(inst.m_Dataset);
  }

  /**
   * Inserts an attribute at the given position (0 to 
   * numAttributes()). Only succeeds if the instance does not
   * have access to any dataset because otherwise inconsistencies
   * could be introduced.
   *
   * @param pos the attribute's position
   * @exception Exception if the instance has accesss to a
   * dataset, or the position is out of range
   */
  public void insertAttributeAt(int position) 
       throws Exception {

    if (m_Dataset != null) {
      throw new Exception("Instance has accesss to a dataset!");
    }
    if ((position < 0) ||
	(position > numAttributes())) {
      throw new Exception("Can't insert attribute: index out "+
			  "of range");
    }
    forceInsertAttributeAt(position);
  }

  /**
   * Tests if a specific value is "missing".
   *
   * @param attIndex the attribute's index
   */
  public boolean isMissing(int attIndex) {

    if (Double.isNaN(m_AttValues[attIndex])) {
      return true;
    }
    return false;
  }

  /**
   * Tests if a specific value is "missing".
   * The given attribute has to belong to a dataset.
   *
   * @param att the attribute
   */
  public boolean isMissing(Attribute att) {

    return isMissing(att.index());
  }

  /**
   * Tests if the given value codes "missing".
   *
   * @param val the value to be tested
   * @return true if val codes "missing"
   */
  public static boolean isMissingValue(double val) {

    return Double.isNaN(val);
  }

  /**
   * Returns the double that codes "missing".
   *
   * @return the double that codes "missing"
   */
  public static double missingValue() {

    return MISSING_VALUE;
  }

  /**
   * Returns the number of attributes.
   *
   * @return the number of attributes as an integer
   */
  public int numAttributes() {

    return m_AttValues.length;
  }

  /**
   * Returns the number of class labels.
   *
   * @return the number of class labels as an integer if the 
   * class attribute is nominal, 1 otherwise.
   * @exception Exception if instance doesn't have access to any
   * dataset
   */
  public int numClasses() throws Exception {
    
    if (m_Dataset == null) {
      throw new Exception("Instance doesn't have access to a dataset!");
    }
    return m_Dataset.numClasses();
  }

  /** 
   * Replaces all missing values in the instance with the modes 
   * and means contained in the given array. A deep copy of
   * the vector of attribute values is performed before the
   * values are replaced.
   *
   * @param array containing the means and modes
   * @exception Exception if numbers of attributes are unequal
   */
  public void replaceMissingValues(double[] array) 
       throws Exception {
	 
    if ((array == null) || 
	(array.length != m_AttValues.length)) {
      throw new Exception("Unequal number of attributes!");
    }
    freshAttributeVector();
    for (int i = 0; i < m_AttValues.length; i++) {
      if (isMissing(i)) {
	m_AttValues[i] = array[i];
      }
    }
  }

  /**
   * Sets the class value of an instance to be "missing". A deep copy of
   * the vector of attribute values is performed before the
   * value is set to be missing.
   *
   * @exception Exception if the class is not set or the instance doesn't
   * have access to a dataset
   */
  public void setClassMissing() throws Exception {

    if (classIndex() < 0) {
      throw new Exception("Class is not set!");
    }
    freshAttributeVector();
    setMissing(classIndex());
  }

  /**
   * Sets the class value of an instance to the given value (internal
   * floating-point format).  A deep copy of the vector of attribute
   * values is performed before the value is set.
   *
   * @param value the new attribute value (If the corresponding
   * attribute is nominal (or a string) then this is the new value's
   * index as a double).  
   * @exception Exception if the class is not set or the instance doesn't
   * have access to a dataset 
   */
  public void setClassValue(double value) throws Exception {

    if (classIndex() < 0) {
      throw new Exception("Class is not set!");
    }
    freshAttributeVector();
    m_AttValues[classIndex()] = value;
  }

  /**
   * Sets the class value of an instance to the given value. A deep
   * copy of the vector of attribute values is performed before the
   * value is set.
   *
   * @param value the new class value (If the class
   * is a string attribute and the value can't be found,
   * the value is added to the attribute).
   * @exception Exception if the dataset or the class is not set, or the 
   * attribute is not nominal or a string, or the value couldn't 
   * be found for a nominal attribute 
   */
  public void setClassValue(String value) throws Exception {

    if (classIndex() < 0) {
      throw new Exception("Class is not set!");
    }
    setValue(classIndex(), value);
  }

  /**
   * Sets the reference to the dataset. Does not check if the instance
   * is compatible with the dataset. Note: the dataset does not know
   * about this instance. If the structure of the dataset's header
   * gets changed, this instance will not be adjusted automatically.
   *
   * @param instances the reference to the dataset 
   */
  public void setDataset(Instances instances) {
    
    m_Dataset = instances;
  }

  /**
   * Sets a specific value to be "missing". Performs a deep copy
   * of the vector of attribute values before the value is set to
   * be missing.
   *
   * @param attIndex the attribute's index
   */
  public void setMissing(int attIndex) throws Exception {

    freshAttributeVector();
    m_AttValues[attIndex] = MISSING_VALUE;
  }

  /**
   * Sets a specific value to be "missing". Performs a deep copy
   * of the vector of attribute values before the value is set to
   * be missing. The given attribute has to belong to a dataset.
   *
   * @param att the attribute
   */
  public void setMissing(Attribute att) throws Exception {

    setMissing(att.index());
  }

  /**
   * Sets a specific value in the instance to the given value 
   * (internal floating-point format). Performs a deep copy
   * of the vector of attribute values before the value is set.
   *
   * @param attIndex the attribute's index 
   * @param value the new attribute value (If the corresponding
   * attribute is nominal (or a string) then this is the new value's
   * index as a double).  
   */
  public void setValue(int attIndex, double value) throws Exception{
    
    freshAttributeVector();
    m_AttValues[attIndex] = value;
  }
  
  public void setValueNoClass(int attIndex, double value) throws Exception
  {
    if (attIndex < this.classIndex()) {
        this.setValue(attIndex, value);
    } else {
        this.setValue(attIndex+1, value);
    }
  }

  /**
   * Sets a value of a nominal or string attribute to the given
   * value. Performs a deep copy of the vector of attribute values
   * before the value is set.
   *
   * @param attIndex the attribute's index
   * @param value the new attribute value (If the attribute
   * is a string attribute and the value can't be found,
   * the value is added to the attribute).
   * @exception Exception if the dataset is not set, or the 
   * attribute is not nominal or a string, or the value couldn't 
   * be found for a nominal attribute 
   */
  public void setValue(int attIndex, String value) throws Exception {
    
    int valIndex;

    if (m_Dataset == null) {
      throw new Exception("Instance doesn't have access to a dataset!");
    }
    if (!attribute(attIndex).isNominal() &&
	!attribute(attIndex).isString()) {
      throw new Exception("Attribute neither nominal nor string!");
    }
    valIndex = attribute(attIndex).indexOfValue(value);
    if (valIndex == -1) {
      if (attribute(attIndex).isNominal()) {
	throw new Exception("Value not defined for given nominal attribute!");
      } else {
	attribute(attIndex).forceAddValue(value);
	valIndex = attribute(attIndex).indexOfValue(value);
      }
    }
    freshAttributeVector();
    m_AttValues[attIndex] = (double)valIndex;
  }

  /**
   * Sets a specific value in the instance to the given value
   * (internal floating-point format). Performs a deep copy of the
   * vector of attribute values before the value is set.
   * The given attribute has to belong to a dataset.
   *
   * @param att the attribute 
   * @param value the new attribute value (If the corresponding
   * attribute is nominal (or a string) then this is the new value's
   * index as a double).
   */
  public void setValue(Attribute att, double value) throws Exception {

    freshAttributeVector();
    m_AttValues[att.index()] = (double)value;
  }

  /**
   * Sets a value of an nominal or string attribute to the given
   * value. Performs a deep copy of the vector of attribute values
   * before the value is set. The given attribute has to belong to a dataset.
   *
   * @param att the attribute
   * @param value the new attribute value (If the attribute
   * is a string attribute and the value can't be found,
   * the value is added to the attribute).
   * @exception Exception if the the attribute is not nominal or a
   * string, or the value couldn't be found for a nominal attribute 
   */
  public void setValue(Attribute att, String value) throws Exception {

    if (!att.isNominal() &&
	!att.isString()) {
      throw new Exception("Attribute neither nominal nor string!");
    }
    int valIndex = att.indexOfValue(value);
    if (valIndex == -1) {
      if (att.isNominal()) {
	throw new Exception("Value not defined for given nominal attribute!");
      } else {
	att.forceAddValue(value);
	valIndex = att.indexOfValue(value);
      }
    }
    freshAttributeVector();
    m_AttValues[att.index()] = (double)valIndex;
  }

  /**
   * Sets the weight of an instance.
   *
   * @param weight the weight
   */
  public void setWeight(double weight) {

    m_Weight = weight;
  }

  /** 
   * Returns the value of a nominal (or string) attribute
   * for the instance.
   *
   * @param attIndex the attribute's index
   * @return the value as a string
   * @exception Exception if the attribute is not a nominal
   * (or string) attribute, or the instance doesn't belong
   * to a dataset.
   */
  public String stringValue(int attIndex) throws Exception {

    if (m_Dataset == null) {
      throw new Exception("Instance doesn't have access to a dataset!");
    } 
    if (!m_Dataset.attribute(attIndex).isNominal() &&
	!m_Dataset.attribute(attIndex).isString()) {
      throw new Exception("Attribute neither nominal nor string!");
    }
    return m_Dataset.attribute(attIndex).
      value((int) m_AttValues[attIndex]);
  }

  /** 
   * Returns the value of a nominal (or string) attribute
   * for the instance.
   *
   * @param att the attribute
   * @return the value as a string
   * @exception Exception if the attribute is not a nominal
   * (or string) attribute, or the instance doesn't belong
   * to a dataset.
   */
  public String stringValue(Attribute att) throws Exception {

    return stringValue(att.index());
  }

  /**
   * Returns the description of one instance. If the instance
   * doesn't have access to a dataset, it returns the internal
   * floating-point values. Quotes string
   * values that contain whitespace characters.
   *
   * @return the instance's description as a string
   */
  public String toString() {

    StringBuffer text = new StringBuffer();
    
    for (int i = 0; i < m_AttValues.length; i++) {
      if (i > 0) text.append(",");
      text.append(toString(i));
    }

    return text.toString();
  }

  /**
   * Returns the description of one value of the instance as a 
   * string. If the instance doesn't have access to a dataset, it 
   * returns the internal floating-point value. Quotes string
   * values that contain whitespace characters, or if they
   * are a question mark.
   *
   * @param attIndex the attribute's index
   * @return the value's description as a string
   */
  public String toString(int attIndex) {

   StringBuffer text = new StringBuffer();
   
   if (isMissing(attIndex)) {
     text.append("?");
   } else {
     if (m_Dataset == null) {
       text.append(Utils.doubleToString(m_AttValues[attIndex],6));
     } else {
       if (m_Dataset.attribute(attIndex).isNominal() || 
	   m_Dataset.attribute(attIndex).isString()) {
	 try {
	   text.append(Utils.quote(stringValue(attIndex)));
	 } catch (Exception e) {
	   throw new Error("This should never happen!");
	 }
       } else {
	 text.append(Utils.doubleToString(m_AttValues[attIndex],6));
       }
     }
   }
   return text.toString();
  }

  /**
   * Returns the description of one value of the instance as a 
   * string. If the instance doesn't have access to a dataset it 
   * returns the internal floating-point value. Quotes string
   * values that contain whitespace characters, or if they
   * are a question mark.
   * The given attribute has to belong to a dataset.
   *
   * @param att the attribute
   * @return the value's description as a string
   */
  public String toString(Attribute att) {
   
   return toString(att.index());
  }

  /**
   * Returns an instance's attribute value in internal format.
   *
   * @param attIndex the attribute's index
   * @return the specified value as a double (If the corresponding
   * attribute is nominal (or a string) then it returns the value's index as a 
   * double).
   */
  public double value(int attIndex) {

    return m_AttValues[attIndex];
  }
    
  public double valueNoClass(int attIndex) throws Exception {
    if (attIndex < this.classIndex()) {
        return this.value(attIndex);        
    } else {
        return this.value(attIndex+1);        
    }
  }
  /**
   * Returns an instance's attribute value in internal format.
   * The given attribute has to belong to a dataset.
   *
   * @param att the attribute
   * @return the specified value as a double (If the corresponding
   * attribute is nominal (or a string) then it returns the value's index as a
   * double).
   */
  public double value(Attribute att) {

    return m_AttValues[att.index()];
  }

  /**
   * Returns the instance's weight.
   *
   * @return the instance's weight as a double
   */
  public double weight(){

    return m_Weight;
  }

  /**
   * Deletes an attribute at the given position (0 to 
   * numAttributes() - 1).
   *
   * @param pos the attribute's position
   */

  void forceDeleteAttributeAt(int position) {

    double[] newValues = new double[m_AttValues.length - 1];

    System.arraycopy(m_AttValues, 0, newValues, 0, position);
    if (position < m_AttValues.length - 1) {
      System.arraycopy(m_AttValues, position + 1, 
		       newValues, position, 
		       m_AttValues.length - (position + 1));
    }
    m_AttValues = newValues;
  }

  /**
   * Inserts an attribute at the given position
   *
   * (0 to numAttributes()) and sets its value to be missing. 
   * @param pos the attribute's position
   */
  void forceInsertAttributeAt(int position)  {

    double[] newValues = new double[m_AttValues.length + 1];

    System.arraycopy(m_AttValues, 0, newValues, 0, position);
    newValues[position] = MISSING_VALUE;
    System.arraycopy(m_AttValues, position, newValues, 
		     position + 1, m_AttValues.length - position);
    m_AttValues = newValues;
  }

  /**
   * Clones the attribute vector of the instance and
   * overwrites it with the clone.
   */
  private void freshAttributeVector() {

    double[] newValues;

    newValues = new double[m_AttValues.length];
    System.arraycopy(m_AttValues, 0, newValues, 0, 
		     m_AttValues.length);
    m_AttValues = newValues;
  }

  /**
   * Main method for testing this class.
   */
  public static void main(String[] options) {

    try {

      // Create numeric attributes "length" and "weight"
      Attribute length = new Attribute("length");
      Attribute weight = new Attribute("weight");
      
      // Create vector to hold nominal values "first", "second", "third" 
      FastVector my_nominal_values = new FastVector(3); 
      my_nominal_values.addElement("first"); 
      my_nominal_values.addElement("second"); 
      my_nominal_values.addElement("third"); 
      
      // Create nominal attribute "position" 
      Attribute position = new Attribute("position", my_nominal_values);
      
      // Create vector of the above attributes 
      FastVector attributes = new FastVector(3);
      attributes.addElement(length);
      attributes.addElement(weight);
      attributes.addElement(position);
      
      // Create the empty dataset "race" with above attributes
      Instances race = new Instances("race", attributes, 0);
      
      // Make position the class attribute
      race.setClassIndex(position.index());
      
      // Create empty instance with three attribute values
      Instance inst = new Instance(3);
      
      // Set instance's values for the attributes "length", "weight", and "position"
      inst.setValue(length, 5.3);
      inst.setValue(weight, 300);
      inst.setValue(position, "first");
      
      // Set instance's dataset to be the dataset "race"
      inst.setDataset(race);
      
      // Print the instance
      System.out.println("The instance: " + inst);
      
      // Print the first attribute
      System.out.println("First attribute: " + inst.attribute(0));
      
      // Print the class attribute
      System.out.println("Class attribute: " + inst.classAttribute());
      
      // Print the class index
      System.out.println("Class index: " + inst.classIndex());
      
      // Say if class is missing
      System.out.println("Class is missing: " + inst.classIsMissing());
      
      // Print the instance's class value in internal format
      System.out.println("Class value (internal format): " + inst.classValue());
      
      // Print a shallow copy of this instance
      Instance copy = (Instance) inst.copy();
      System.out.println("Shallow copy: " + copy);
      
      // Set dataset for shallow copy
      copy.setDataset(inst.dataset());
      System.out.println("Shallow copy with dataset set: " + copy);
      
      // Unset dataset for copy, delete first attribute, and insert it again
      copy.setDataset(null);
      copy.deleteAttributeAt(0);
      copy.insertAttributeAt(0);
      copy.setDataset(inst.dataset());
      System.out.println("Copy with first attribute deleted and inserted: " + copy); 
      
      // Enumerate attributes (leaving out the class attribute)
      System.out.println("Enumerating attributes (leaving out class):");
      Enumeration enum = inst.enumerateAttributes();
      while (enum.hasMoreElements()) {
	Attribute att = (Attribute) enum.nextElement();
	System.out.println(att);
      }
      
      // Headers are equivalent?
      System.out.println("Header of original and copy equivalent: " +
			 inst.equalHeaders(copy));

      // Test for missing values
      System.out.println("Length of copy missing: " + copy.isMissing(length));
      System.out.println("Weight of copy missing: " + copy.isMissing(weight.index()));
      System.out.println("Length of copy missing: " + 
			 Instance.isMissingValue(copy.value(length)));
      System.out.println("Missing value coded as: " + Instance.missingValue());

      // Prints number of attributes and classes
      System.out.println("Number of attributes: " + copy.numAttributes());
      System.out.println("Number of classes: " + copy.numClasses());

      // Replace missing values
      double[] meansAndModes = {2, 3, 0};
      copy.replaceMissingValues(meansAndModes);
      System.out.println("Copy with missing value replaced: " + copy);

      // Setting and getting values and weights
      copy.setClassMissing();
      System.out.println("Copy with missing class: " + copy);
      copy.setClassValue(0);
      System.out.println("Copy with class value set to first value: " + copy);
      copy.setClassValue("third");
      System.out.println("Copy with class value set to \"third\": " + copy);
      copy.setMissing(1);
      System.out.println("Copy with second attribute set to be missing: " + copy);
      copy.setMissing(length);
      System.out.println("Copy with length set to be missing: " + copy);
      copy.setValue(0, 0);
      System.out.println("Copy with first attribute set to 0: " + copy);
      copy.setValue(weight, 1);
      System.out.println("Copy with weight attribute set to 1: " + copy);
      copy.setValue(position, "second");
      System.out.println("Copy with position set to \"second\": " + copy);
      copy.setValue(2, "first");
      System.out.println("Copy with last attribute set to \"first\": " + copy);
      System.out.println("Current weight of instance copy: " + copy.weight());
      copy.setWeight(2);
      System.out.println("Current weight of instance copy (set to 2): " + copy.weight());
      System.out.println("Last value of copy: " + copy.toString(2));
      System.out.println("Value of position for copy: " + copy.toString(position));
      System.out.println("Last value of copy (internal format): " + copy.value(2));
      System.out.println("Value of position for copy (internal format): " + 
			 copy.value(position));
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
  
  public void setIndex(int indx)
  {
    this.m_index = indx;
  }
  
  public int getIndex()
  {
    return this.m_index;
  }
  
  public Instance toInstance(Instances dataSet) throws Exception
  {
    Instance copyIns = this.makeCopy();
    if (dataSet.numAttributes() != copyIns.numAttributes()) {
        throw new Exception("number of attributes is different");        
    }
    if (dataSet.classIndex() != copyIns.classIndex()) {
        throw new Exception("class index is different");
    }
    copyIns.setDataset(dataSet);
    return copyIns;
  }
  
  public FastVector ids() throws Exception
  {
      int i;
      FastVector ids = new FastVector();
      for (i=0; i<this.m_AttValues.length; i++) {
        if (isMissingValue(this.m_AttValues[i])) {
            throw new Exception("Missing values not suported");
        }
        ids.addElement(new Integer(i));        
      }
      return ids;
  }
    
  public FastVector values() throws Exception
  {
      int i;    
      FastVector vals = new FastVector();
      for (i=0; i<this.m_AttValues.length; i++) {
        if (isMissingValue(this.m_AttValues[i])) {
            throw new Exception("Missing values not suported");
        }
        vals.addElement(new Double(this.m_AttValues[i]));        
      }      
      return vals;
  }
}
